﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using EonaCat.Helpers.Controls;

namespace EonaCat.Sockets
{
    public class SocketServer : IDisposable
    {
        /// <summary>
        /// Unique server id.
        /// </summary>
        public string Guid { get; set; }

        /// <summary>
        /// Listening ip address.
        /// </summary>
        public IPAddress IPAddress { get; set; }

        /// <summary>
        /// Listening port.
        /// </summary>
        public int Port { get; set; }

        /// <summary>
        /// List of connected clients.
        /// </summary>
        public List<Client> ConnectedClients { get; set; } = new List<Client>();

        /// <summary>
        /// Interval to check for disconnected clients
        /// </summary>
        public int DisconnectionCheckInterval { get; set; } = 100;

        /// <summary>
        /// Determines if we need to reuse the socket address
        /// (default: false)
        /// </summary>
        public bool ReuseAddress { get; set; }

        /// <summary>
        /// Determines if the socket is in blocking mode
        /// </summary>
        public bool IsSocketBlockingDisabled { get; set; }


        /// <summary>
        /// Socket backlog. The maximum length of the pending connections queue.
        /// </summary>
        public int Backlog { get; set; } = 0;
        /// <summary>
        /// Server type.
        /// </summary>
        public NetworkType ServerType { get; set; }

        /// <summary>
        /// Only for UDP.
        /// (default: true)
        /// </summary>
        public bool UseClientManager { get; set; } = true;

        /// <summary>
        /// Only for UDP.
        /// Accept 1 byte data for connected state check.
        /// </summary>
        private int _UDPDataInterval { get => (int)(UDPDataInterval * 1.5); }
        public int UDPDataInterval { get; set; } = 5000;

        /// <summary>
        /// Called when new client connected
        /// </summary>
        public event EventHandler<SocketServerClientEventArgs> OnConnected;
        /// <summary>
        /// Called when client disconnected
        /// </summary>
        public event EventHandler<SocketServerClientEventArgs> OnDisconnected;
        /// <summary>
        /// Called when received data from client
        /// </summary>
        public event EventHandler<SocketServerDataEventArgs> OnReceived;
        /// <summary>
        /// Called when server catch error
        /// </summary>
        public event EventHandler<ErrorEventArgs> OnError;
        /// <summary>
        /// Called when executing Start() function
        /// </summary>
        public event EventHandler<string> OnStart;
        /// <summary>
        /// Called when executing Stop() function
        /// </summary>
        public event EventHandler OnStop;

        private Socket Server { get; set; }
        private EonaCatTimer DisconnectTimerUdp { get; set; }
        private EonaCatTimer DisconnectTimerTcp { get; set; }
        private Dictionary<EndPoint, double> LastDataReceivedTime { get; set; }
        private double TimeNow { get => (DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds; }

        public SocketServer(NetworkType type, string address, int port)
        {
            if (!IPAddress.TryParse(address, out var ipAddress))
            {
                throw new Exception("Invalid IP address specified");
            }

            SetupSocketServer(type, ipAddress, port);
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="type">Server type (TCP, UDP)</param>
        /// <param name="address">Server listening ip</param>
        /// <param name="port">Server listening port</param>
        public SocketServer(NetworkType type, IPAddress address, int port)
        {
            SetupSocketServer(type, address, port);
        }

        private void SetupSocketServer(NetworkType type, IPAddress address, int port)
        {
            Guid = System.Guid.NewGuid().ToString();
            ServerType = type;
            IPAddress = address;
            Port = port;
        }

        private readonly List<EndPoint> _removedEndpoints = new List<EndPoint>();

        private void UdpCheckCallback()
        {
            _removedEndpoints.Clear();
            var now = TimeNow;
            foreach (var kp in LastDataReceivedTime.Where(kp => now - kp.Value > _UDPDataInterval / 1000))
            {
                lock (ConnectedClients)
                {
                    var client = ConnectedClients.FirstOrDefault(x => x.Guid == kp.Key.ToString());
                    if (client == null)
                    {
                        continue;
                    }
                    OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                    ConnectedClients.Remove(client);
                    _removedEndpoints.Add(kp.Key);
                }
            }

            lock (LastDataReceivedTime)
            {
                foreach (var r in _removedEndpoints)
                {
                    LastDataReceivedTime.Remove(r);
                }
            }
        }

        private void TcpCheckCallback()
        {
            lock (ConnectedClients)
            {
                ConnectedClients.RemoveAll(x =>
                {
                    if (x.IsConnected)
                    {
                        return false;
                    }

                    OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = x });
                    return true;
                });
            }
        }

        public void Dispose()
        {
            Stop();
        }

        /// <summary>
        /// Start server
        /// </summary>
        public async Task StartAsync()
        {
            try
            {
                switch (ServerType)
                {
                    case NetworkType.Tcp:
                        Server = new Socket(IPAddress.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
                        ConfigureSocketOptions();

                        // Check if we need to override the socket with a custom socket
                        if (CustomTcpSocket != null)
                        {
                            Server = CustomTcpSocket;
                        }

                        DisconnectTimerTcp = new EonaCatTimer(TimeSpan.FromMilliseconds(DisconnectionCheckInterval), TcpCheckCallback);
                        DisconnectTimerTcp.Start();
                        break;
                    case NetworkType.Udp:
                        {
                            if (UseClientManager)
                            {
                                LastDataReceivedTime = new Dictionary<EndPoint, double>();
                                DisconnectTimerUdp = new EonaCatTimer(TimeSpan.FromMilliseconds(DisconnectionCheckInterval), UdpCheckCallback);
                                DisconnectTimerUdp.Start();
                            }
                            Server = new Socket(IPAddress.AddressFamily, SocketType.Dgram, ProtocolType.Udp);
                            ConfigureSocketOptions();

                            // Check if we need to override the socket with a custom socket
                            if (CustomUdpSocket != null)
                            {
                                Server = CustomUdpSocket;
                            }
                            break;
                        }
                }
                Server.ReceiveBufferSize = Client.BufferSize;
                Server.SendBufferSize = Client.BufferSize;
                Server.Bind(new IPEndPoint(IPAddress, Port));
                if (ServerType == NetworkType.Tcp)
                {
                    Server.Listen(Backlog);
                }
                await Task.Run(ListenLoop);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        /// <summary>
        /// Setup this udp socket to override the default one
        /// (default: null => using default)
        /// </summary>
        public Socket CustomUdpSocket { get; set; }

        /// <summary>
        /// Setup this tcp socket to override the default one
        /// (default: null => using default)
        /// </summary>
        public Socket CustomTcpSocket { get; set; }

        static void SetConnectionReset(Socket socket)
        {
            try
            {
                if (socket.ProtocolType == ProtocolType.Udp && RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
                {
                    // Disable ICMP packet shutdown (forcibly closed)
                    const uint IOC_IN = 0x80000000;
                    const uint IOC_VENDOR = 0x18000000;
                    uint SIO_UDP_CONNRESET = IOC_IN | IOC_VENDOR | 12;
                    socket.IOControl((int)SIO_UDP_CONNRESET, new[] { Convert.ToByte(false) }, null);
                };
            }
            catch { }
        }

        private void ConfigureSocketOptions()
        {
            // Disable ICMP packet shutdown
            SetConnectionReset(Server);

            if (ReuseAddress)
            {
                Server.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
            }

            Server.Blocking = IsSocketBlockingDisabled;
        }

        /// <summary>
        /// Stop server
        /// </summary>
        public void Stop()
        {
            try
            {
                DisconnectTimerUdp?.Stop();
                DisconnectTimerTcp?.Stop();
                foreach (var client in ConnectedClients)
                {
                    OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                    client.Socket.Close();
                }
                ConnectedClients.Clear();
                OnStop?.Invoke(this, EventArgs.Empty);
                Server.Close();
            }
            catch (SocketException socketException)
            {
                OnError?.Invoke(this, new ErrorEventArgs(socketException));
            }
        }

        private void ListenLoop()
        {
            try
            {
                OnStart?.Invoke(this, CustomTcpSocket != null || CustomUdpSocket != null ? "Custom socket was used" : string.Empty);

                switch (ServerType)
                {
                    case NetworkType.Tcp:
                        Server.BeginAccept(AcceptCallback, Server);
                        break;

                    case NetworkType.Udp:
                        var loopbackAddress = Server.AddressFamily == AddressFamily.InterNetworkV6
                            ? IPAddress.IPv6Loopback
                            : IPAddress.Loopback;

                        var buffer = new byte[Client.BufferSize];
                        EndPoint remoteIp = new IPEndPoint(loopbackAddress, 0);
                        Server.BeginReceiveFrom(buffer, 0, Client.BufferSize, 0, ref remoteIp, UseClientManager ? AcceptClientsCallbackUDP : ReadClientCallbackUdp, new object[] { buffer, remoteIp, Server });
                        break;
                    default:
                        throw new NotSupportedException("Protocol type not supported");
                }
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        private void AcceptCallback(IAsyncResult ar)
        {
            try
            {
                var listener = (Socket)ar.AsyncState;
                var handler = listener.EndAccept(ar);
                var client = new Client()
                {
                    Guid = System.Guid.NewGuid().ToString(),
                    Socket = handler
                };
                lock (ConnectedClients)
                {
                    ConnectedClients.Add(client);
                }
                OnConnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                client.Socket.BeginReceive(client.Buffer, 0, Client.BufferSize, 0, ReadCallback, client);
                listener.BeginAccept(AcceptCallback, Server);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void ReadCallback(IAsyncResult ar)
        {
            try
            {
                var client = (Client)ar.AsyncState;
                var bytesRead = client.Socket.EndReceive(ar);
                if (bytesRead > 0)
                {
                    var data = new byte[bytesRead];
                    Array.Copy(client.Buffer, 0, data, 0, bytesRead);
                    OnReceived?.Invoke(this, new SocketServerDataEventArgs() { Client = client, Data = data });
                }
                client.Socket.BeginReceive(client.Buffer, 0, Client.BufferSize, 0, ReadCallback, client);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void AcceptClientsCallbackUDP(IAsyncResult ar)
        {
            try
            {
                var objects = (object[])ar.AsyncState;
                var buffer = objects[0] as byte[];
                var remoteIp = objects[1] as EndPoint;
                var listener = objects[2] as Socket;
                var bytesRead = listener.EndReceiveFrom(ar, ref remoteIp);

                Client client;
                lock (ConnectedClients)
                {
                    client = ConnectedClients.FirstOrDefault(x => x.Guid == remoteIp.ToString());
                    if (client == null)
                    {
                        client = new Client()
                        {
                            Guid = remoteIp.ToString(),
                            Socket = new Socket(remoteIp.AddressFamily, SocketType.Dgram, ProtocolType.Udp)
                        };
                        client.SendData += SendData;
                        client.Socket.Connect(remoteIp);
                        if (client.IsConnected)
                        {
                            ConnectedClients.Add(client);
                            OnConnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                        }
                    }
                }

                if (bytesRead > 1)
                {
                    var data = new byte[bytesRead];
                    Array.Copy(buffer, 0, data, 0, bytesRead);
                    OnReceived?.Invoke(this, new SocketServerDataEventArgs() { Client = client, Data = data, });
                }
                else
                {
                    lock (LastDataReceivedTime)
                    {
                        if (LastDataReceivedTime.ContainsKey(remoteIp))
                        {
                            LastDataReceivedTime[remoteIp] = TimeNow;
                        }
                        else
                        {
                            LastDataReceivedTime.Add(remoteIp, TimeNow);
                        }
                    }
                }

                Server.BeginReceiveFrom(buffer, 0, Client.BufferSize, 0, ref remoteIp, AcceptClientsCallbackUDP, new object[] { buffer, remoteIp, listener });
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void ReadClientCallbackUdp(IAsyncResult ar)
        {
            try
            {
                var objects = (object[])ar.AsyncState;
                var buffer = objects[0] as byte[];
                var remoteIp = objects[1] as EndPoint;
                var listener = objects[2] as Socket;
                var bytesRead = listener.EndReceiveFrom(ar, ref remoteIp);
                if (bytesRead > 0)
                {
                    var data = new byte[bytesRead];
                    Array.Copy(buffer, 0, data, 0, bytesRead);
                    OnReceived?.Invoke(this, new SocketServerDataEventArgs()
                    {
                        Data = data,
                        Client = new Client()
                        {
                            Guid = remoteIp.ToString(),
                            Socket = new Socket(remoteIp.AddressFamily, SocketType.Dgram, ProtocolType.Udp)
                        }
                    });
                }
                Server.BeginReceiveFrom(buffer, 0, Client.BufferSize, 0, ref remoteIp, ReadClientCallbackUdp, new object[] { buffer, remoteIp, Server });
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
            catch (ObjectDisposedException) { }
        }

        private void DisconnectCallback(IAsyncResult ar)
        {
            try
            {
                var client = (Client)ar.AsyncState;
                client.Socket.EndDisconnect(ar);
                OnDisconnected?.Invoke(this, new SocketServerClientEventArgs() { Client = client });
                lock (ConnectedClients)
                {
                    ConnectedClients.Remove(client);
                }
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        public void SendData(byte[] data, EndPoint point)
        {
            try
            {
                Server.BeginSendTo(data, 0, data.Length, 0, point, EndSendTo, null);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }

        private void EndSendTo(IAsyncResult ar)
        {
            try
            {
                Server.EndSendTo(ar);
            }
            catch (SocketException se)
            {
                OnError?.Invoke(this, new ErrorEventArgs(se));
            }
        }
    }
}